@@ -6,7 +6,7 @@ use warnings;
 
 use Perinci::Sub::Wrapper qw(wrap_sub);
 
-our $VERSION = '0.05'; # VERSION
+our $VERSION = '0.06'; # VERSION
 
 use Moose;
 use experimental 'smartmatch';
@@ -41,7 +41,7 @@ has _wrap_args_compiled => (
 has _prereqs => (
     is      => 'rw',
 );
-has _checked_modules => (
+has _registered_modules => (
     is      => 'rw',
     default => sub { {} },
 );
@@ -51,7 +51,8 @@ sub _squish_code {
     my ($self, $code) = @_;
     return $code if $self->debug;
     for ($code) {
-        s/^\s*#.+//mg; # comment line
+        s/^\s*#.+//mg; # comment line on its own
+        s/##.*//mg;    # trailing comment using ##
         s/^\s+//mg;    # indentation
         s/\n+/ /g;     # newline
     }
@@ -148,7 +149,7 @@ sub munge_file {
         my $src = $res->[2]{source};
         for (split /^/, $src->{presub1}) {
             push @requires, $_ unless $_ ~~ @requires;
-            if (/^\s*(?:use|require) \s+ (\w+(?:::\w+)*)/x) {
+            if (/^\s*(?:use|require|no) \s+ (\w+(?:::\w+)*)/x) {
                 $mods{$1}++;
             }
         }
@@ -156,48 +157,20 @@ sub munge_file {
 
     return unless keys %wres;
 
-    # check dist.ini to make sure that non-core modules required by the wrapper
-    # code is mentioned in Prereqs. XXX this should use proper method. XXX we
-    # should not parse dist.ini for each file.
+    # register prereqs for validator code
     {
-
-        unless ($self->_prereqs) {
-            require Module::CoreList;
-
-            (-f "dist.ini") or do {
-                $self->log_fatal("dist.ini not found, something's wrong");
-            };
-            my $in_prereqs;
-            my %prereqs;
-            $self->log_debug("Parsing prereqs in dist.ini ...");
-            open my($fh), "<", "dist.ini";
-            while (<$fh>) {
-                next unless /\S/;
-                next if /\s*;/;
-                if (/^\s*\[\s*([^\]+]+)\s*\]/) {
-                    $in_prereqs = $1 eq 'Prereqs';
-                    next;
-                }
-                next unless $in_prereqs;
-                /^\s*(\S+)\s*=\s*(\S+)/ or
-                    $self->log_fatal("dist.ini:$.: syntax error");
-                $prereqs{$1} = $2;
-            }
-            #use Data::Dump; dd \%prereqs;
-            $self->_prereqs(\%prereqs);
-        }
-
-        my $perl_version = $self->_prereqs->{perl};
-
+        require Module::CoreList;
         for my $mod (sort keys %mods) {
-            next if Module::CoreList::is_core($mod, undef, $perl_version);
-            next if $self->_checked_modules->{$mod};
-            $self->log_debug("Checking if dist.ini contains Prereqs for $mod");
-            $self->log_fatal(
-                "Wrapper code requires non-core module '$mod', ".
-                    "please specify it in dist.ini's Prereqs section")
-                unless defined $self->_prereqs->{$mod};
-            $self->_checked_modules->{$mod}++;
+            next if Module::CoreList::is_core($mod, undef);
+            next if $self->_registered_modules->{$mod};
+            if ($self->zilla->prereqs->cpan_meta_prereqs->{requirements}{$mod}) {
+                $self->log_debug("Prereq for validator code has already been specified in dist.ini: $mod");
+                next;
+            }
+            $self->log("Adding prereq for validator code: $mod");
+            $self->zilla->register_prereqs(
+                {phase=>'runtime'}, $mod => 0);
+            $self->_registered_modules->{$mod}++;
         }
     }
 
@@ -206,8 +179,9 @@ sub munge_file {
     my $sub_name; # current subname
     my $sub_indent;
 
-    my $has_put_preamble;
+    my $has_preamble;
     my $has_postamble;
+    my $has_put_preamble;
     my $has_put_postamble;
 
   LINE:
@@ -223,21 +197,39 @@ sub munge_file {
             next;
         }
 
-        if (/^(\s*)sub \s+ (\w+)/x) {
+        if (/^(\s*)sub \s+ (\w+)(?: \s* \{ \s* (\#\s*NO_RINCI_WRAP) )?/x) {
+            my $no_wrap;
             $self->log_debug("Found sub declaration: $2");
             my $first_sub = !$sub_name;
-            ($sub_indent, $sub_name) = ($1, $2);
+            ($sub_indent, $sub_name, $no_wrap) = ($1, $2, $3);
+
+            # XXX last sub doesn't get this check
+            if ($has_postamble && !$has_put_postamble) {
+                $self->log_fatal("[sub $sub_name] hasn't put postamble ".
+                                     "wrapper code yet");
+            }
+
             unless ($wres{$sub_name}) {
                 $self->log_debug("Skipped wrapping sub $sub_name (no metadata)");
+                $sub_name = undef;
+                next;
+            }
+            if ($no_wrap) {
+                $self->log_debug("Skipped wrapping sub $sub_name (#NO_RINCI_WRAP directive)");
+                $sub_name = undef;
                 next;
             }
             # put modify-meta code
-            $_ = "\n$1# [Rinci::Wrap] END presub2\n$_" if $self->debug;
-            $_ = $self->_squish_code($wres{$sub_name}{source}{presub2}). " $_";
-            $_ = "\n$1# [Rinci::Wrap] BEGIN presub2\n$_" if $self->debug;
-            $has_put_preamble  = 0 || !$wres{$sub_name}{source}{preamble};
-            $has_postamble     = $wres{$sub_name}{source}{postamble} ? 1:0;
-            $has_put_postamble = 0 || !$has_postamble;
+            my $presub2 = $self->_squish_code($wres{$sub_name}{source}{presub2});
+            if ($presub2 =~ /\S/) {
+                $_ = "\n$1# [Rinci::Wrap] END presub2\n$_" if $self->debug;
+                $_ = "$presub2 $_";
+                $_ = "\n$1# [Rinci::Wrap] BEGIN presub2\n$_" if $self->debug;
+            }
+            $has_preamble      = $wres{$sub_name}{source}{preamble} =~ /\S/;
+            $has_postamble     = $wres{$sub_name}{source}{postamble} =~ /\S/;
+            $has_put_preamble  = 0;
+            $has_put_postamble = 0;
 
             if ($first_sub) {
                 # this is the first sub, let's put all requires here
@@ -253,17 +245,27 @@ sub munge_file {
 
         # 'my %args = @_' statement
         if (/^(\s*)(my \s+ [\%\@\$]args \s* = .+)/x) {
-            $self->log_debug("[sub $sub_name] Found a place to insert preamble (after '$2' statement)");
-            # put preamble code
-            $_ = "$_\n$1# [Rinci::Wrap] BEGIN preamble\n" if $self->debug;
-            my $preamble = $wres{$sub_name}{source}{preamble};
-            if ($has_postamble) {
-                $preamble .= $1 . '$_w_res = do {';
+            {
+                last unless $has_preamble && !$has_put_preamble;
+                $self->log_debug("[sub $sub_name] Found a place to insert preamble (after '$2' statement)");
+
+                my $indent = $1;
+
+                # remove comment that might interfere with preamble adding
+                s/(.+)#.*/$1/;
+
+                # put preamble code
+                $_ = "$_\n$indent# [Rinci::Wrap] BEGIN preamble\n" if $self->debug;
+                my $preamble = $wres{$sub_name}{source}{preamble};
+                if ($has_postamble) {
+                    $preamble .= $indent . '$_w_res = do {';
+                }
+                s/\n//;
+                $_ .= " " . $self->_squish_code($preamble) . "\n";
+                $_ = "$_\n$indent# [Rinci::Wrap] END preamble\n" if $self->debug;
+                $has_put_preamble = 1;
+                next LINE;
             }
-            $_ .= " " . $self->_squish_code($preamble);
-            $_ = "$_\n$1# [Rinci::Wrap] END preamble\n" if $self->debug;
-            $has_put_preamble = 1;
-            next;
         }
 
         # sub closing statement
@@ -271,35 +273,31 @@ sub munge_file {
             $self->log_debug("Found sub closing: $sub_name");
             next unless $wres{$sub_name};
 
-            unless ($has_put_preamble) {
-                $self->log_fatal("[sub $sub_name] hasn't put preamble ".
-                                     "wrapper code yet");
+            {
+                if ($has_preamble && !$has_put_preamble) {
+                    $self->log_fatal("[sub $sub_name] hasn't put preamble ".
+                                         "wrapper code yet");
+                }
+                last unless $has_postamble && !$has_put_postamble;
+
+                # put postamble code
+                my $postamble = "}; " . # for closing of the do { block
+                    $wres{$sub_name}{source}{postamble};
+                $_ = "\n$sub_indent# [Rinci::Wrap] END postamble\n$_"
+                    if $self->debug;
+                $_ = $self->_squish_code($postamble) . " $_";
+                $_ = "\n$sub_indent# [Rinci::Wrap] BEGIN postamble\n$_"
+                    if $self->debug;
+                $has_put_postamble = 1;
+
             }
-            goto DONE_POSTAMBLE unless $has_postamble;
-
-            # put postamble code
-            my $postamble = "}; " . # for closing of the do { block
-                $wres{$sub_name}{source}{postamble};
-            $_ = "\n$sub_indent# [Rinci::Wrap] END postamble\n$_"
-                if $self->debug;
-            $_ = $self->_squish_code($postamble) . " $_";
-            $_ = "\n$sub_indent# [Rinci::Wrap] BEGIN postamble\n$_"
-                if $self->debug;
-            $has_put_postamble = 1;
-
-          DONE_POSTAMBLE:
             # mark sub done by deleting entry from %wres
             delete $wres{$sub_name};
 
-            next;
+            next LINE;
         }
     }
 
-    if (!$has_put_postamble && $sub_name) {
-        $self->log_fatal("[sub $sub_name] hasn't put postamble ".
-                             "wrapper code yet");
-    }
-
     if (keys %wres) {
         $self->log_fatal("Some subs are not yet wrapped (probably because I couldn't find sub declaration or a place to insert the preamble/postamble): ".
                              join(", ", sort keys %wres));
@@ -326,7 +324,7 @@ Dist::Zilla::Plugin::Rinci::Wrap - Insert wrapper-generated code
 
 =head1 VERSION
 
-version 0.05
+This document describes version 0.06 of Dist::Zilla::Plugin::Rinci::Wrap (from Perl distribution Dist-Zilla-Plugin-Rinci-Wrap), released on 2015-02-23.
 
 =head1 SYNOPSIS
 
@@ -426,7 +424,7 @@ your code if you want to use this plugin.
 The generated wrapper code will declare some variables. You need to make sure
 that the variables do not clash. This is rather simple: the variables used by
 the wrapper code will all be prefixed with C<_w_> (e.g. C<$_w_res>) or C<_sahv_>
-for variables generated by the L<Sah> schema compiler (e.g. <$_sahv_dpath>).
+for variables generated by the L<Sah> schema compiler (e.g. C<$_sahv_dpath>).
 
 =item * Variable used to accept arguments
 
@@ -449,11 +447,11 @@ If you accept arguments from a hashref (i.e. C<< func({ arg1=>1, arg2=>2 }) >>):
 
  my $args = $_[0] // {};
 
-If you accept arguments from an array (e.g. C<<func(1, 2)>>:
+If you accept arguments from an array (e.g. C<< func(1, 2) >>:
 
  my @args = @_;
 
-If you accept arguments from an arrayref C<<func([1, 2])>>:
+If you accept arguments from an arrayref C<< func([1, 2]) >>:
 
  my $args = $_[0] // [];
 
@@ -486,36 +484,20 @@ behaviours/functionalities.
 =head2 But the generated code looks ugly!
 
 Admittedly, yes. Wrapper-generated code is formatted as a single long line to
-avoid modifying line numbers, which is desirable when debugging your modules. An
-option to not compress everything as a single line might be added in the future.
+avoid modifying line numbers, which is desirable when debugging your modules. If
+you don't want to compress everything as a single line, add C<debug=1> in your
+C<dist.ini>.
 
 =head2 How do I customize wrapping for my function
 
-In the future there will be options you can specify in C<dist.ini>.
-
-For now, you can put in your Rinci function metadata:
+Two ways. You can either use C<wrap_args> in C<dist.ini> (see Synopsis) or add
+an attribute in your function metadata:
 
  "x.dist.zilla.plugin.rinci.wrap.wrap_args" => { validate_args => 0 },
 
-This will be merged and will override C<wrap_args> keys specified in
+which will be merged and will override C<wrap_args> keys specified in
 C<dist.ini>.
 
-=head1 TODO
-
-=over
-
-=item * Use L<PPI> instead of fragile regex.
-
-=item * Option to not compress wrapper code to a single line.
-
-=item * Option to reuse validation code for the same schema.
-
-=item * Option to exclude some subroutines from being wrapped.
-
-=item * Option to specify different wrap_args for different subroutines.
-
-=back
-
 =head1 HOMEPAGE
 
 Please visit the project's homepage at L<https://metacpan.org/release/Dist-Zilla-Plugin-Rinci-Wrap>.
@@ -534,11 +516,11 @@ feature.
 
 =head1 AUTHOR
 
-Steven Haryanto <stevenharyanto@gmail.com>
+perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2014 by Steven Haryanto.
+This software is copyright (c) 2015 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
